// This file was generated by ecm_setup_version(): DO NOT EDIT!

#ifndef KDB_VERSION_H
#define KDB_VERSION_H

#define KDB_VERSION_STRING "3.2.0"
#define KDB_VERSION_MAJOR 3
#define KDB_VERSION_MINOR 2
#define KDB_VERSION_PATCH 0
#define KDB_VERSION ((3<<16)|(2<<8)|(0))

#endif
